#include "majong.h"

char *tile_chr[] = {
    "1c ", "2c ", "3c ", "4c ", "5c ", "6c ", "7c ", "8c ", "9c ",
    "1d ", "2d ", "3d ", "4d ", "5d ", "6d ", "7d ", "8d ", "9d ",
    "1b ", "2b ", "3b ", "4b ", "5b ", "6b ", "7b ", "8b ", "9b ",
    "T ", "N ", "S ", "P ", "W ", "G ", "R "};

char honor_chrs[] = {'t', 'n', 's', 'p', 'w', 'g', 'r'};

int opt_chin;
int opt_time;
int opt_brute;
int brute_level = 2;
t_melds last_max_melds;
int last_mentsu_shanten;
int last_seventwins_shanten;
int last_kokushi_shanten;

/*
 *
 *  HumanInterface Routines
 *
 */

int main(int argc,char *argv[])
{
    printf("welcome to shanten program!\n");
    printf("type q to quit, h to view help\n");
    while(1){
        printf("mj : ");
        char buff[255];
        fgets(buff, 255, stdin);
        int len = strlen(buff) - 1;
        buff[len] = 0; //cut off LF
        if( 0 == strcmp( buff, "random") || 0 == strcmp( buff, "r") || 0 == len){
            printf("randomhaipai!\n");
            int tiles[13];
            int res = random_haipai(tiles, 34);
            try_and_print(tiles, res);
        }
        else if( 0 == strcmp( buff, "brute") ){
            opt_brute = !opt_brute;
            if(opt_brute){
                printf("level? : ");
                fgets(buff, 255, stdin);
                len = strlen(buff) - 1;
                buff[len] = 0; //cut off LF
                brute_level = atoi(buff);
                printf("enter the brute force mode level %d!\n", brute_level);
                printf("level 3 or more deeper brute force search is DEADLY HEAVY. take care.\n");
            }else{
                printf("exit from brute fore mode!\n");
            }
        }
        else if( 0 == strcmp( buff, "chin") || 0 == strcmp( buff, "c") ){
            opt_chin = !opt_chin;
            printf("%s\n", opt_chin ? "enter the chin-ichi mode!" : "exit from chin-ichi mode!");
        }
        else if( 0 == strcmp( buff, "help") || 0 == strcmp( buff, "h") ){
            print_usage();
        }
        else if( 0 == strcmp( buff, "quit") || 0 == strcmp( buff, "q") ){
            printf("bye\n");
            break;
        }
        else if( 0 == strcmp( buff, "bench") || 0 == strcmp( buff, "b") ){
            printf("times? : ");
            fgets(buff, 255, stdin);
            len = strlen(buff) - 1;
            buff[len] = 0; //cut off LF
            int repeat_times = atoi(buff);
            benchmark(repeat_times);
        }
        else {
            int tiles[13];
            int res;
            if(res = str2tilearray(buff, tiles, 14)){
                try_and_print(tiles, res);
            }else{
                printf("what?\n");
            }
        }
    }
}

void try_and_print(int *tiles, int tile_count){
    int s = shanten(tiles, tile_count);
    t_melds m = last_max_melds;
    char str[255];
    tilearray2str(tiles, tile_count, str, 255);
    printf("--------\n");
    printf("checking for %d tiles...\n", tile_count);
    printf("sorted_tehai  : %s\n", str);
    printf("best mentsu : \n");
    printf("  mentsu | tahtsu | toitsu : %d | %d | %d\n", m.mentsu, m.tahtsu, m.toitsu);
    printf("  float suit  | dead suit  : %d | %d\n", m.float_suit_tiles, m.dead_suit_tiles);
    printf("  float honor | dead honot : %d | %d\n", m.float_honor_tiles, m.dead_honor_tiles);
    printf("each shanten:\n");
    printf("  mentsu | seventwins | kokushi : %d | %d | %d\n", last_mentsu_shanten, last_seventwins_shanten, last_kokushi_shanten);
    printf("shanten-suu : %d\n", s); 

    if(s == 0 && tile_count == 13){
        int waiting_tiles[34];
        int i, waiting_count = 0;
        for (i = 0; i<34; i++){
            if (can_win_with(tiles, 13, i)) {
                waiting_tiles[waiting_count++] = i;
            }
        }
        if(!waiting_count){ 
            printf("something error occured in core routines\n");
        }else{
            char waiting_str[256];
            tilearray2str(waiting_tiles, waiting_count, waiting_str, 256);
            printf("waiting for : %s\n", waiting_str);
        }
    }
    if(opt_brute){
        int b_result = brute_force_shanten_search(tiles);
        if( b_result == -1){
            printf("there is no winning path in %d level brute force search...\n", brute_level);
        }else{
            printf("brute force result : %d\n", b_result);
        }
    }
}

void benchmark(int repeat_times){
    int i, sum = 0, mentsu_sum = 0, seventwins_sum = 0, kokushi_sum = 0;
    clock_t t1,t2;

    t1 = clock();

    for(i=0;i<repeat_times;i++){
        int tiles[13];
        int res;
        res = random_haipai(tiles, 34);
        sum += shanten(tiles,res);
        mentsu_sum += last_mentsu_shanten;
        seventwins_sum += last_seventwins_shanten;
        kokushi_sum += last_kokushi_shanten;
    }
    if(last_mentsu_shanten < 0) printf("out mentsn\n");
    if(last_seventwins_shanten < 0) printf("out seventwins\n");
    if(last_kokushi_shanten < 0) printf("out kokushi\n");
    t2 = clock();

    double tt = (double)(t2 - t1) / (double)CLOCKS_PER_SEC;
    printf("time = %fs for %d times loop. proc %f/s\n", tt, repeat_times, repeat_times/tt);
    printf("           avg = %f\n", (double)sum / (double)repeat_times);
    printf("mentsu     avg = %f\n", (double)mentsu_sum / (double)repeat_times);
    printf("seventwins avg = %f\n", (double)seventwins_sum / (double)repeat_times);
    printf("kokushi    avg = %f\n", (double)kokushi_sum / (double)repeat_times);
}

int compare_int(const int *a, const int *b){ return *a - *b; }

int random_haipai(int *buf, int limit){
    int deck[136];
    int i;
    for(i=0;i<136;i++) deck[i] = 1;
    for(i=0;i<13;i++) buf[i] = 0;
    for(i=0;i<13;){
        int rnd = rand() % (opt_chin ? 36 : 136);
        if(deck[rnd]){
            buf[i] = rnd / 4;
            deck[rnd]--;
            i++;
        }
    }
    qsort(buf, 13, sizeof(int), (int (*)(const void*, const void*))compare_int);
    return i;
}

void print_usage() {
    printf("random | r          : get random tiles and calculate shanten-suu of that hand.\n");
    printf("(enter w/o command) : one more alias of random command.\n");
    printf("bench | b           : try benchmark with random hand. enter the number of times to try.\n");
    printf("                      if you are in the chin-ichi mode, be careful to try with the big num.\n");
    printf("help | h            : view this help.\n");
    printf("chin | c            : toggle the chin-ichi mode on and off.\n");
    printf("                      in chin-ichi mode, use one suit only random hand, in random and bench command.\n");
    printf("quit | q            : quit this program.\n");
    printf("(your hand)         : type your hand, and you get an result of the calculation.\n");
    printf("                      here is markup rules;\n");
    printf("                      \n");
    printf("suit tiles:\n");
    printf("    basically, a tile will be marked-up by a pair of the rank number and suit character.\n");
    printf("        1b 2b 3b 5c 5c 5c 2d 5d 8d\n");
    printf("    here are suit characters:\n");
    printf("        b : bamboos\n");
    printf("        c : characters\n");
    printf("        d : dots\n");
    printf("    if the tiles are same suit, you can use block style. see below:\n");
    printf("        123b 555c 258d\n");
    printf("honor tiles:\n");
    printf("    just type an initial character.\n");
    printf("        T : Tong (east)\n");
    printf("        N : Nang (south)\n");
    printf("        S : Sha (west)\n");
    printf("        P : Pei (north)\n");
    printf("        W : White\n");
    printf("        G : Green\n");
    printf("        R : Red\n");
}


/*
 *
 *  Utility Routines
 *
 */

int tilearray2str(int *tiles, int tile_count, char *buf, int buf_size){
    int i=0,pos=0;
    for(i=0;i<tile_count;i++){
        char *chr = tile_chr[tiles[i]];
        strcpy(buf + pos, chr);
        pos += strlen(chr);
    }
    buf[pos] = 0;
    return 1;
}

int honorchr2tilenum(char c){
    int i=0;
    for(i=0;i<7;i++){
        if(honor_chrs[i] == c) return i + 27;
    }
    return 0;
}

int str2tilearray(char *str, int *buf, int buf_size){
    int pos, i=0, honor=0, suit_base = 0;
    for(pos = strlen(str) - 1; pos>=0; pos--){
        char c = str[pos];
        if('A' <= c && c <= 'Z'){
            c -= 0x20;
        }
        if(c == ' '){
            continue;
        }else if(c == 'c'){
            suit_base = -1;
        }else if(c == 'd'){
            suit_base = 8;
        }else if(c == 'b'){
            suit_base = 17;
        }else if('1' <= c && c <= '9'){
            if(!suit_base) return 0;
            buf[i++] = suit_base + (c - '0');
        }else if( honor = honorchr2tilenum(c) ){
            buf[i++] = honor;
        }else return 0;
    }
    qsort(buf, i, sizeof(int), (int (*)(const void*, const void*))compare_int);
    return i;
}

