#include "majong.h"

int shanten(int *tiles, int tile_count){
    int i;
    int tile_abstract[34];
    for(i=0;i<34;i++)
        tile_abstract[i] = 0;
    for(i=0;i<tile_count;i++)
        tile_abstract[tiles[i]]++;
    int mentsu_shanten = count_mentsu_shanten(tile_abstract);
    int seventwins_shanten = count_seventwins_shanten(tile_abstract);
    int kokushi_shanten = count_kokushi_shanten(tile_abstract);

    last_mentsu_shanten = mentsu_shanten;
    last_seventwins_shanten = seventwins_shanten;
    last_kokushi_shanten = kokushi_shanten;

    int min;
    min = mentsu_shanten < seventwins_shanten ? mentsu_shanten : seventwins_shanten;
    min = min < kokushi_shanten ? min : kokushi_shanten;
    return min;
}

int count_mentsu_shanten(int *tiles){
    t_melds melds = count_max_melds(tiles);

    int shanten = 8;
    shanten -= melds.mentsu * 2;
    if(melds.tahtsu > 5 - melds.mentsu){
        shanten -= 5 - melds.mentsu;
    }else if(melds.tahtsu == 5 - melds.mentsu){
        shanten -= melds.tahtsu;
        if(!melds.toitsu) shanten++;
    } else {
        shanten -= melds.tahtsu;
    }

    int dead = melds.dead_suit_tiles + melds.dead_honor_tiles;
    if(melds.toitsu && melds.dead_suit_tiles) dead--;
    shanten = dead > shanten ? dead : shanten;
    return shanten;
}

int count_seventwins_shanten (int *tiles){
    int i, twins = 0, overplus = 0;
    for(i=0;i<34;i++){
        if(tiles[i] >= 2) twins++;
        if(tiles[i] > 2) overplus += tiles[i] - 2;
    }
    return (6 - twins) > overplus ? 6 - twins : overplus;
}

int count_kokushi_shanten (int *tiles) {
    int i,count=0, head=0;
    int targets[] = {0,8,9,17,18,26,27,28,29,30,31,32,33};
    for(i=0;i<13;i++){
        if(tiles[targets[i]]) count++;
        if(tiles[targets[i]] >= 2) head++;
    }
    return head ? 12 - count : 13 - count;
}

t_melds count_max_melds(int *tiles){
    t_melds total;
    total.mentsu            = 0;
    total.tahtsu            = 0;
    total.toitsu            = 0;
    total.float_suit_tiles  = 0;
    total.dead_suit_tiles   = 0;
    total.float_honor_tiles = 0;
    total.dead_honor_tiles  = 0;

    int i;
    t_melds melds_in_suit[4];

    melds_in_suit[0] = max_available_melds_in_suit(tiles     );
    melds_in_suit[1] = max_available_melds_in_suit(tiles + 9 );
    melds_in_suit[2] = max_available_melds_in_suit(tiles + 18);
    melds_in_suit[3] = max_available_melds_in_honor(tiles + 27);

    for(i=0;i<4;i++){
        total.mentsu            += melds_in_suit[i].mentsu;
        total.tahtsu            += melds_in_suit[i].tahtsu;
        total.toitsu            += melds_in_suit[i].toitsu;
        total.float_suit_tiles  += melds_in_suit[i].float_suit_tiles;
        total.dead_suit_tiles   += melds_in_suit[i].dead_suit_tiles;
        total.float_honor_tiles += melds_in_suit[i].float_honor_tiles;
        total.dead_honor_tiles  += melds_in_suit[i].dead_honor_tiles;
    }
    last_max_melds = total;
    return total;
}

t_melds max_available_melds_in_suit(int *tiles){
    int i, tile_count = 0;
    for(i=0;i<9;i++) tile_count += tiles[i];

    int max = get_max_meld_comparer_in_suit_fast(tiles, 0, 0);
    t_melds melds;
    melds.mentsu            = (max & 0xF000) >> 12;
    melds.tahtsu            = (max & 0x0F00) >> 8;
    melds.toitsu            = (max & 0x00F0) >> 4;
    melds.float_suit_tiles  = max & 0x000F;
    melds.dead_suit_tiles   = 0;
    melds.float_honor_tiles = 0;
    melds.dead_honor_tiles  = 0;


    melds.dead_suit_tiles = tile_count 
                          - melds.mentsu * 3
                          - melds.tahtsu * 2
                          - melds.float_suit_tiles;
    return melds;
}

t_melds max_available_melds_in_honor(int *tiles){
    t_melds melds;
    melds.mentsu            = 0;
    melds.tahtsu            = 0;
    melds.toitsu            = 0;
    melds.float_suit_tiles  = 0;
    melds.dead_suit_tiles   = 0;
    melds.float_honor_tiles = 0;
    melds.dead_honor_tiles  = 0;

    int i;
    for(i=0;i<7;i++){
        int cnt = tiles[i];
        if(cnt == 4){
            melds.mentsu++;
            melds.dead_honor_tiles++;
        }
        else if(cnt == 3){
            melds.mentsu++;
        }
        else if(cnt == 2){
            melds.tahtsu++;
            melds.toitsu++;
        }
        else if(cnt == 1){
            melds.float_honor_tiles++;
        }
    }
    return melds;
}

int get_max_meld_comparer_in_suit_fast(int *tiles, int base, int repeater){

    if(base > 8) return 0;
    while(!(tiles[base])){
        base++;
        repeater = 0;
        if(base > 8) return 0;
    }

    int max = 0;
    int count;

    tiles[base]--;

    if (base < 7 && tiles[base + 2]){
        tiles[base + 2]--;
        count = get_max_meld_comparer_in_suit_fast(tiles, base, 1);
        count += 0x0100;
        max = max > count ? max : count;

        if (tiles[base + 1]){
            tiles[base + 1]--;
            count = get_max_meld_comparer_in_suit_fast(tiles, base, 1);
            count += 0x1000;
            max = max > count ? max : count;
            tiles[base + 1]++;
        }
        tiles[base + 2]++;
    }

    if (base < 8 && tiles[base + 1]){
        tiles[base + 1]--;
        count = get_max_meld_comparer_in_suit_fast(tiles, base, 1);
        count += 0x0100;
        max = max > count ? max : count;
        tiles[base + 1]++;
    }

    if(repeater){
        tiles[base]++;
        return max;
    }

    if(tiles[base] <= 2){
        count = get_max_meld_comparer_in_suit_fast(tiles, base, 1);
        count += 0x0001;
        max = max > count ? max : count;
    }

    if (tiles[base]){
        tiles[base]--;
        count = get_max_meld_comparer_in_suit_fast(tiles, base, 1);
        count += 0x0110;
        max = max > count ? max : count;

        if (tiles[base]){
            tiles[base]--;
            count = get_max_meld_comparer_in_suit_fast(tiles, base, 1);
            count += 0x1000;
            max = max > count ? max : count;

            if (tiles[base]){
                tiles[base]--;
                count = get_max_meld_comparer_in_suit_fast(tiles, base + 1, 0);
                count += 0x1000;
                max = max > count ? max : count;
                tiles[base]++;
            }
            tiles[base]++;
        }
        tiles[base]++;
    }

    tiles[base]++;
    return max;
}

int can_win_with(int *tiles, int tile_count, int win_tile) {
    if (tile_count != 13) return 0;
    int i;
    int tile_abstract[34];
    for (i=0; i<34; i++) tile_abstract[i] = 0;
    for (i=0; i<13; i++) tile_abstract[tiles[i]]++;
    tile_abstract[win_tile]++;

    if(win_mentsu(tile_abstract)) return 1;
    if(win_seventwins(tile_abstract)) return 1;
    if(win_kokushi(tile_abstract)) return 1;
    return 0;
}

int win(int *tiles, int tile_count) {
    if (tile_count != 14) return 0;
    int i;
    int tile_abstract[34];
    for (i=0;i<34;i++) tile_abstract[i] = 0;
    for (i=0;i<14;i++) tile_abstract[tiles[i]]++;
    if(win_mentsu(tile_abstract)) return 1;
    if(win_seventwins(tile_abstract)) return 1;
    if(win_kokushi(tile_abstract)) return 1;
    return 0;
}

int win_mentsu(int *tiles) {
    int head[7];
    int i, head_count = 0;
    for(i=0;i<34;i++){
        if(tiles[i] >= 2) head[head_count++] = i;
    }
    if(!head_count) return 0;
    for (i=0;i<head_count;i++) {
        tiles[head[i]] -= 2;
        int mentsu = get_all_mentsu(tiles);
        tiles[head[i]] += 2;
        if (mentsu == 4) return 1;
    }
    return 0;
}

int get_all_mentsu(int *tiles) {

    int i, base = -1, count = 0;
    for (i=0;i<34;i++){
        if(tiles[i]) { base = i; break; }
    }
    if (base == -1) return 0;

    if (tiles[base] >= 3) {
        tiles[base] -= 3;
        count = get_all_mentsu(tiles);
        tiles[base] += 3;
        if(count >= 0) return count + 1;
    }

    if (base <= 24 && (base % 9) < 7) {
        if( tiles[base + 1] && tiles[base + 2] ){
            tiles[base]--;
            tiles[base + 1]--;
            tiles[base + 2]--;
            count = get_all_mentsu(tiles);
            tiles[base]++;
            tiles[base + 1]++;
            tiles[base + 2]++;
            if(count >= 0) return count + 1;
        }
    }
    return -1;
}

int win_seventwins(int *tiles) {
    int i, twins = 0;
    for(i=0;i<34;i++){
        if(tiles[i] == 2) twins++;
    }
    return (twins == 7);
}

int win_kokushi(int *tiles) {
    int i,head=0;
    int targets[] = {0,8,9,17,18,26,27,28,29,30,31,32,33};
    for(i=0;i<13;i++){
        if(!tiles[targets[i]]) return 0;
        if(tiles[targets[i]] >= 2) head++;
    }
    return head;
}

static int deck_yama[34];
static int brute_max_depth;

int brute_force_tsumo(int *tiles, int depth)
{
    int i,j;
    if (depth == brute_max_depth) {
        for (i=0; i<34; i++) {
            if ( deck_yama[i]>=4 ) continue;
            if(can_win_with(tiles, 13, i)) return 1;
        }
        return 0;
    }

    for (i=0; i<34; i++) {
        if ( deck_yama[i]>=4 ) continue;
        deck_yama[i]++;
        
        for (j=0;j<13;j++) {
            int copy = tiles[j];
            tiles[j] = i;
            if ( brute_force_tsumo(tiles, depth + 1) ) return 1;
            tiles[j] = copy;
        }
        deck_yama[i]--;
    }
    return 0;
}

int brute_force_shanten_search(int *tiles)
{
    int i;
    for (i=0;i<34;i++) deck_yama[i] = 0;    
    for (i=0;i<13;i++) deck_yama[tiles[i]]++;
    
    int brute_result = -1;    
        
    for (i = 0; i <= brute_level; i++) {
        brute_max_depth = i;
        if ( brute_force_tsumo(tiles, 0) ) {
            brute_result = brute_max_depth;
            break;
        }
    }

    return brute_result;
}
